/* minimization of L/L#-convex function
 *  This is an implementation of steepest descent algorithm
 *  with submodular set function minimize method,
 *  and its scaling algorithm.
 *  cf. K. Murota: Discrete convex analysis (Japanese), 2001, p.240-243.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "odicon.h"

/* returns  0 if no improvment was found. */
/* returns +1 if improvment was found in plus set */
/* returns -1 if improvment was found in minus set */
static int best_improve(int dim, double func(int,int*), int x[],
                        int width, int *best_set, const int fix_index) {
    int improved = 0; /* -1,0,1 */
    int dir;
    int set[1000];
    double f, min_f;

    iteration++;
    min_f = 0;
    for (dir=-1; dir<=1; dir+=2) { /* dir = {-1, 1} */
        f = sf_minimize(dim, func, x, dir*width, set, fix_index);
        if (f< min_f) {
            min_f = f;  improved = dir;  improve_call++;
            ivector_copy(dim+1, best_set, set);
        }
    }
    return improved;
}


static void steepest_descent(int dim, double func(int,int*), int x[],
                             const int width, const int fix_index) {
    int i, dir, set[1000];
    int init_iteration = iteration;

    while ( (dir = best_improve(dim, func, x, width, set, fix_index)) != 0 ) {
        for (i=0; i<set[0]; i++) x[set[i+1]] += width * dir;
#ifdef DEBUG
        if (width > 1) {
            printf("width=%d, step=%d, ", width, iteration - init_iteration);
        }
#endif /* DEBUG */
	loop_print(dim, func, x);
    }

    if (width > 1 && iteration - init_iteration > dim) {
        printf("#iteration = %d > dim\n", iteration - init_iteration);
    }
}

static double minimize(int dim, double func(int,int*), int init[],
                       int width, const int fix_index, enum submodular_e type) {
    submodular_type = type;
    counter_reset();
    if (!is_feasible(dim, func, init)) return PLUS_INFINITY;
    status_print(dim, func, init);

    while (width > 1) {
        steepest_descent(dim, func, init, width, fix_index);
        width /= 2 * sqrt(dim);  /* imploved by N.T */
    }
    steepest_descent(dim, func, init, 1, fix_index);

    counter_println();
    minimizer_print(dim, func, init);
    return func(dim, init);
}


static double minimize_scaling(int dim, double func(int,int*), int init[],
                               const int fix_index, enum submodular_e type) {
    int width = 20 * sqrt(dim);
    return minimize(dim, func, init, width, fix_index, type);
}

/* returns the minimum value of func(x)    */
/* INPUT  func   : pointer to the function */
/* INPUT  init[] : the initial solution    */
/* OUTPUT init[] : the minimized solution  */
double lgconv_minimize_IFF(int dim, double func(int,int*), int init[]) {
    return minimize(dim, func, init, 1, -1, IFF);
}
double lgconv_minimize_FW(int dim, double func(int,int*), int init[]) {
    return minimize(dim, func, init, 1, -1, FW);
}

double lgconv_minimize_scaling_IFF(int dim, double func(int,int*), int init[]) {
    return minimize_scaling(dim, func, init, -1, IFF);
}

double lgconv_minimize_scaling_FW(int dim, double func(int,int*), int init[]) {
    return minimize_scaling(dim, func, init, -1, FW);
}

double lconv_minimize_IFF(int dim, double func(int,int*), int init[],
                          const int fix_index) {
    return minimize(dim, func, init, 1, fix_index, IFF);
}
double lconv_minimize_FW(int dim, double func(int,int*), int init[],
                         const int fix_index) {
    return minimize(dim, func, init, 1, fix_index, FW);
}
double lconv_minimize_scaling_IFF(int dim, double func(int,int*), int init[],
                                  const int fix_index) {
    return minimize_scaling(dim, func, init, fix_index, IFF);
}
double lconv_minimize_scaling_FW(int dim, double func(int,int*), int init[],
                                 const int fix_index) {
    return minimize_scaling(dim, func, init, fix_index, FW);
}
